#ifndef YSSEGARRAY_IS_INCLUDED
#define YSSEGARRAY_IS_INCLUDED
/* { */

#include "ysdef.h"
#include "ysarray.h"

/*! A variable length array class.
    This class stores array items as a segmented sub-arrays.  Since those sub-arrays are not
    continuous, this class cannot be casted to an ordinary C array.
    
    Since the sub-arrays will not be re-located like doubling-array, the location of the 
    array items will not change.  Therefore, it is safer to cache a pointer to an array
    item.
    */

template <class T,const unsigned int BitShift>
class YsSegmentedArray
{
// \cond
private:
	enum
	{
		SEGMENT_UNIT_SIZE=(1<<BitShift)
	};

	YSSIZE_T nv;
	YsArray <T *> segments;

protected:
	inline YSRESULT Alloc(YSSIZE_T n);
// \endcond

public:
	/*! Default constructor */
	inline YsSegmentedArray();

	/*! Copy constructor */
	inline YsSegmentedArray(const YsSegmentedArray <T,BitShift> &incoming);

	/*! Copy constructor */
	template <const unsigned int BitShift2>
	inline YsSegmentedArray(const YsSegmentedArray <T,BitShift2> &incoming);

	/*! Copy constructor */
	inline YsSegmentedArray(YSSIZE_T nv,const T v[]);

	/*! Default destructor */
	inline ~YsSegmentedArray();

	/*! This function copies v[0] to v[nv-1] to this array. */
	inline YSRESULT Set(YSSIZE_T nv,const T v[]);

	/*! This function copies incoming segmented array to this array. */
	inline YSRESULT CopyFrom(const YsSegmentedArray <T,BitShift> &incoming);

	/*! This function copies incoming segmented array to this array. */
	template <const unsigned int BitShift2>
	inline YSRESULT CopyFrom(const YsSegmentedArray <T,BitShift2> &incoming);

	/*! Copy operator */
	inline const YsSegmentedArray <T,BitShift> &operator=(const YsSegmentedArray <T,BitShift> &incoming);

	/*! Copy operator */
	template <const unsigned int BitShift2>
	inline const YsSegmentedArray <T,BitShift> &operator=(const YsSegmentedArray <T,BitShift2> &incoming);

	/*! Returns number of items stored in this array. */
	inline YSSIZE_T GetN(void) const;
	/*! Returns the total size of the array allocated for this array.  This may be bigger than GetN() */
	inline YSSIZE_T NAvailable(void) const;
	/*! This function return YSTRUE if 0<=id && id<GetN() */
	inline YSBOOL IsInRange(YSSIZE_T id) const;  // 2008/07/06
	/*! Access operator */
	inline T &operator[](YSSIZE_T idx);
	/*! Access operator */
	inline const T &operator[](YSSIZE_T idx) const;

	/*! Set number of items to zero, but does not free allocated sub-arrays. */
	void Clear(void);
	/*! Set number of items to zero and free all allocated sub-arrays. */
	void ClearDeep(void);

	/*! This function adds an item to the end of the array. */
	inline YSRESULT Append(const T &dat);

	/*! This function adds multiple items given by nItm and itm to the end of the array. 
	   \param nDat [In] Number of items to be added.
	   \param dat [In] Items to be added.
	   */
	inline YSRESULT Append(YSSIZE_T nDat,const T dat[]);

	/*! This function adds multiple items given by ary to the end of the array. 
	   */
	template <const int BitShift2>
	inline YSRESULT Append(const YsSegmentedArray <T,BitShift2> &incoming);

	/*! This function allocates the buffer good enough for the newSize.  It changes
	    NAvailable(), but does not change GetN(). */
	inline YSRESULT Resize(YSSIZE_T newSize);

	/*! This function deletes unused sub-arrays. */
	inline YSRESULT Shrink(void);

	/*! This function increases the array size by one. */
	inline YSRESULT Increment(void);

	/*! This function inserts dat to the array.  dat will become idx-th item after this
	    function. */
	inline YSRESULT Insert(YSSIZE_T idx,const T &dat);

	/*! This function deletes idx-th item.  Latter items will be shifted (the order doesn't change). */
	inline YSRESULT Delete(YSSIZE_T idx);

	/*! This function copies the last item of the array to the idx-th item, and shrinks
	    the length of the array by one.  When the order of the items in the array does not
	    matter, this function quickly deletes idx-th item in the array.
	    \param idx [In] Index to the item to be deleted.
	   */
	inline YSRESULT DeleteBySwapping(YSSIZE_T idx);  // Added 2001/04/27

	/*! This function deletes the last item of the array. */
	inline YSRESULT DeleteLast(void);

	/*!  */
	inline T &GetTop(void);
	/*!  */
	inline const T& GetTop(void) const;
	/*!  */
	inline T &GetEnd(void);
	/*!  */
	inline const T& GetEnd(void) const;
};

template <class T,const unsigned int BitShift>
inline YSRESULT YsSegmentedArray <T,BitShift>::Alloc(YSSIZE_T n)
{
	YSSIZE_T reqUnit=(n+SEGMENT_UNIT_SIZE-1)/SEGMENT_UNIT_SIZE;
	if(segments.GetN()<reqUnit)
	{
		YSSIZE_T n0=segments.GetN();
		if(YSOK==segments.Resize(reqUnit))
		{
			for(YSSIZE_T i=n0; i<segments.GetN(); i++)
			{
				segments[i]=new T [SEGMENT_UNIT_SIZE];
			}
			return YSOK;
		}
	}
	else if(reqUnit+1<segments.GetN())
	{
		for(YSSIZE_T i=reqUnit+1; i<segments.GetN(); i++)
		{
			delete [] segments[i];
		}
		return segments.Resize(reqUnit+1);
	}
	return YSOK;
}

template <class T,const unsigned int BitShift>
inline YsSegmentedArray <T,BitShift>::YsSegmentedArray()
{
	nv=0;
}


template <class T,const unsigned int BitShift>
inline YsSegmentedArray <T,BitShift>::YsSegmentedArray(const YsSegmentedArray <T,BitShift> &incoming)
{
	CopyFrom(incoming);
}


template <class T,const unsigned int BitShift>
template <const unsigned int BitShift2>
inline YsSegmentedArray <T,BitShift>::YsSegmentedArray(const YsSegmentedArray <T,BitShift2> &incoming)
{
	CopyFrom(incoming);
}


template <class T,const unsigned int BitShift>
inline YsSegmentedArray <T,BitShift>::YsSegmentedArray(YSSIZE_T nv,const T v[])
{
	Set(nv,v);
}


template <class T,const unsigned int BitShift>
inline YsSegmentedArray <T,BitShift>::~YsSegmentedArray()
{
	ClearDeep();
}

template <class T,const unsigned int BitShift>
inline YSRESULT YsSegmentedArray <T,BitShift>::Set(YSSIZE_T nv,const T v[])
{
	if(0==nv)
	{
		Clear();
		return YSOK;
	}
	else
	{
		if(YSOK==Resize(nv))
		{
			for(YSSIZE_T i=0; i<nv; i++)
			{
				(*this)[i]=v[i];
			}
			return YSOK;
		}
	}
	return YSERR;
}

template <class T,const unsigned int BitShift>
inline YSRESULT YsSegmentedArray <T,BitShift>::CopyFrom(const YsSegmentedArray <T,BitShift> &incoming)
{
	if(YSOK==Resize(incoming.GetN()))
	{
		for(YSSIZE_T i=0; i<incoming.GetN(); i++)
		{
			(*this)[i]=incoming[i];
		}
		return YSOK;
	}
	return YSERR;
}


template <class T,const unsigned int BitShift>
template <const unsigned int BitShift2>
inline YSRESULT YsSegmentedArray <T,BitShift>::CopyFrom(const YsSegmentedArray <T,BitShift2> &incoming)
{
	if(YSOK==Resize(incoming.GetN()))
	{
		for(YSSIZE_T i=0; i<incoming.GetN(); i++)
		{
			(*this)[i]=incoming[i];
		}
		return YSOK;
	}
	return YSERR;
}



template <class T,const unsigned int BitShift>
inline const YsSegmentedArray <T,BitShift> &YsSegmentedArray <T,BitShift>::operator=(const YsSegmentedArray <T,BitShift> &incoming)
{
	CopyFrom(incoming);
	return *this;
}


template <class T,const unsigned int BitShift>
template <const unsigned int BitShift2>
inline const YsSegmentedArray <T,BitShift> &YsSegmentedArray <T,BitShift>::operator=(const YsSegmentedArray <T,BitShift2> &incoming)
{
	CopyFrom(incoming);
	return *this;
}

template <class T,const unsigned int BitShift>
inline YSSIZE_T YsSegmentedArray <T,BitShift>::GetN(void) const
{
	return nv;
}


template <class T,const unsigned int BitShift>
inline YSSIZE_T YsSegmentedArray <T,BitShift>::NAvailable(void) const
{
	return SEGMENT_UNIT_SIZE*segments.GetN();
}

template <class T,const unsigned int BitShift>
inline YSBOOL YsSegmentedArray <T,BitShift>::IsInRange(YSSIZE_T id) const
{
	if(0<=id && id<GetN())
	{
		return YSTRUE;
	}
	return YSFALSE;
}

template <class T,const unsigned int BitShift>
inline T &YsSegmentedArray <T,BitShift>::operator[](YSSIZE_T idx)
{
	const YSSIZE_T seg=(idx>>BitShift);
	const YSSIZE_T offset=(idx&((1<<BitShift)-1));
	return segments[seg][offset];
}


template <class T,const unsigned int BitShift>
inline const T &YsSegmentedArray <T,BitShift>::operator[](YSSIZE_T idx) const
{
	const YSSIZE_T seg=(idx>>BitShift);
	const YSSIZE_T offset=(idx&((1<<BitShift)-1));
	return segments[seg][offset];
}

template <class T,const unsigned int BitShift>
void YsSegmentedArray <T,BitShift>::Clear(void)
{
	nv=0;
}

template <class T,const unsigned int BitShift>
void YsSegmentedArray <T,BitShift>::ClearDeep(void)
{
	for(YSSIZE_T i=0; i<segments.GetN(); i++)
	{
		T *ptr=segments[i];
		delete [] ptr;
	}
	segments.ClearDeep();
	nv=0;
}


template <class T,const unsigned int BitShift>
inline YSRESULT YsSegmentedArray <T,BitShift>::Append(const T &dat)
{
	if(YSOK==Increment())
	{
		(*this)[nv-1]=dat;
		return YSOK;
	}
	return YSERR;
}


template <class T,const unsigned int BitShift>
inline YSRESULT YsSegmentedArray <T,BitShift>::Append(YSSIZE_T nDat,const T dat[])
{
	const YSSIZE_T n0=GetN();
	Resize(GetN()+nDat);
	for(YSSIZE_T i=0; i<nDat; i++)
	{
		(*this)[n0+i]=dat[i];
	}
	return YSOK;
}


template <class T,const unsigned int BitShift>
template <const int BitShift2>
inline YSRESULT YsSegmentedArray <T,BitShift>::Append(const YsSegmentedArray <T,BitShift2> &incoming)
{
	const YSSIZE_T n0=GetN();
	Resize(GetN()+incoming.GetN());
	for(YSSIZE_T i=0; i<incoming.GetN(); i++)
	{
		(*this)[n0+i]=incoming[i];
	}
	return YSOK;
}

template <class T,const unsigned int BitShift>
inline YSRESULT YsSegmentedArray <T,BitShift>::Resize(YSSIZE_T newSize)
{
	if(YSOK==Alloc(newSize))
	{
		nv=newSize;
		return YSOK;
	}
	return YSERR;
}

template <class T,const unsigned int BitShift>
inline YSRESULT YsSegmentedArray <T,BitShift>::Shrink(void)
{
	YSSIZE_T reqUnit=(nv+SEGMENT_UNIT_SIZE-1)/SEGMENT_UNIT_SIZE;
	for(YSSIZE_T i=reqUnit; i<segments.GetN(); i++)
	{
		delete [] segments[i];
	}
	segments.Resize(reqUnit);
	return YSOK;
}


template <class T,const unsigned int BitShift>
inline YSRESULT YsSegmentedArray <T,BitShift>::Increment(void)
{
	return Resize(GetN()+1);
}

template <class T,const unsigned int BitShift>
YSRESULT YsSegmentedArray <T,BitShift>::Insert(YSSIZE_T idx,const T &dat)
{
	if(YSOK==Increment())
	{
		for(YSSIZE_T i=GetN()-1; idx<i; i--)
		{
			(*this)[i]=(*this)[i-1];
		}
		return YSOK;
	}
	return YSERR;
}

template <class T,const unsigned int BitShift>
YSRESULT YsSegmentedArray <T,BitShift>::Delete(YSSIZE_T idx)
{
	for(YSSIZE_T i=idx; idx<GetN()-1; idx++)
	{
		(*this)[i]=(*this)[i+1];
	}
	Resize(GetN()-1);
	return YSOK;
}

template <class T,const unsigned int BitShift>
YSRESULT YsSegmentedArray <T,BitShift>::DeleteBySwapping(YSSIZE_T idx)
{
	(*this)[idx]=GetEnd();
	Resize(GetN()-1);
	return YSOK;
}

template <class T,const unsigned int BitShift>
inline YSRESULT YsSegmentedArray <T,BitShift>::DeleteLast(void)
{
	Resize(GetN()-1);
	return YSOK;
}

template <class T,const unsigned int BitShift>
inline T &YsSegmentedArray <T,BitShift>::GetTop(void)
{
	return (*this)[0];
}


template <class T,const unsigned int BitShift>
inline const T& YsSegmentedArray <T,BitShift>::GetTop(void) const
{
	return (*this)[0];
}


template <class T,const unsigned int BitShift>
inline T &YsSegmentedArray <T,BitShift>::GetEnd(void)
{
	return (*this)[nv-1];
}


template <class T,const unsigned int BitShift>
inline const T& YsSegmentedArray <T,BitShift>::GetEnd(void) const
{
	return (*this)[nv-1];
}

/* } */
#endif
